import unittest

import sympy as sp
from sympy.matrices import eye
from tensors.indexed import IndexObject, Manifold, NotContainedSymbolException
from tensors.indexed import Tensor, ChristoffelSymbol


class IndexedTest(unittest.TestCase):

    def test_display(self):
        # Act
        x, y = sp.symbols("x, y")
        vector = IndexObject(r"\phi", (x, y), {(x, x): 1, (y, y): 2})

        actual = vector.display()
        expected = r"$$\phi_x{}_x = 1 \\ \phi_y{}_y = 2$$"
        # Assert
        self.assertEqual(expected, actual)

    def test_get_item(self):
        # Act
        x, y = sp.symbols("x, y")
        vector = IndexObject(r"\phi", (x, y), {(x, x): 1, (y, y): 2})
        actual = vector[x, x]
        # Assert
        self.assertEqual(1, actual)

    def test_display_three(self):
        # Arrange
        x, y, z = sp.symbols("x, y, z")
        vector = IndexObject(r"\phi", (x, y, z), {(x, x, x): 1, (y, y, y): 2, (y, z, z): 3})

        # Act
        actual = vector.display()
        expected = r"$$\phi_x{}_x{}_x = 1 \\ \phi_y{}_y{}_y = 2 \\ \phi_y{}_z{}_z = 3$$"

        # Assert
        self.assertEqual(actual, expected)

    def test_non_given_coordinate(self):
        # Act
        x, y = sp.symbols("x, y")
        vector = IndexObject(r"\phi", (x, y), {(x, x): 1, (y, y): 2})

        actual = vector[x, y]
        # Assert
        self.assertEqual(0, actual)

    def test_non_contained_coordinate(self):
        # Act
        x, y, z = sp.symbols("x, y,z")
        vector = IndexObject(r"\phi", (x, y), {(x, x): 1, (y, y): 2})
        with self.assertRaises(NotContainedSymbolException):
            vector[x, z]

    def test_non_contained_coordinate_constructor(self):
        # Act
        x, y, z = sp.symbols("x, y,z")
        with self.assertRaises(NotContainedSymbolException):
            vector = IndexObject(r"\phi", (x, y), {(x, z): 1, (y, y): 2})

    def test_display_1up2down(self):
        # Arrange
        x, y, z = sp.symbols("x, y, z")
        vector = IndexObject(r"\phi", (x, y, z), {(x, x, x): 1, (y, y, y): 2, (y, z, z): 3}, index_pos=(1, 0, 0))

        # Act
        actual = vector.display()
        expected = r"$$\phi^x{}_x{}_x = 1 \\ \phi^y{}_y{}_y = 2 \\ \phi^y{}_z{}_z = 3$$"

        # Assert
        self.assertEqual(actual, expected)


class ManifoldTest(unittest.TestCase):

    def test_init(self):
        x, y, z = sp.symbols("x, y, z")
        coords = (x, y, z)
        manifold = Manifold(coords,
                            {(x, x): 1, (y, y): 1, (z, z): 1})

        assert manifold.coords == coords
        assert type(manifold.metric) == Tensor

    def test_convert_sympy_matrix(self):
        # Arrange
        x, y, z = sp.symbols("x, y, z")
        coords = (x, y, z)
        manifold = Manifold(coords,
                            {(x, x): 1, (y, y): 1, (z, z): 1})

        # Act
        actual = manifold.metric_to_matrix()
        expected = eye(3)

        # Assert
        self.assertEqual(actual, expected)

    def test_manifold_from_matrix(self):
        # Arrange
        matrix = eye(3)

        # Act
        x, y, z = sp.symbols("x, y, z")
        coords = (x, y, z)
        actual = Manifold(coords, matrix)
        expected = Manifold(coords,
                            {(x, x): 1, (y, y): 1, (z, z): 1})

        # Assert
        assert actual.metric.display() == expected.metric.display()


class TensorTest(unittest.TestCase):
    def test_init(self):
        x, y, z = sp.symbols("x, y, z")
        coords = (x, y, z)
        manifold = Manifold(coords,
                            {(x, x): 1, (y, y): 1, (z, z): 1})
        tensor = Tensor(r"\phi",
                        manifold,
                        (0, 0),
                        {(x, x): 1, (x, y): 2, (z, z): 3})

        assert tensor.coords == coords
        assert tensor.manifold is not None

        actual = tensor[x, y]
        # Assert
        self.assertEqual(2, actual)


class ChristoffelSymbolTest(unittest.TestCase):
    def test_init(self):
        x, y, z = sp.symbols("x, y, z")
        coords = (x, y, z)

        manifold = Manifold(coords,
                            {(x, x): 1, (y, y): 1, (z, z): 1})
        cs = ChristoffelSymbol(manifold)

        assert cs.values == {}

    def test_polar_coords(self):
        rho, theta = sp.symbols(r"\rho, \theta")
        coords = (rho, theta)

        manifold = Manifold(coords,
                            {(rho, rho): 1, (theta, theta): rho ** 2})
        cs = ChristoffelSymbol(manifold)
        print(cs.values)
        true_values={(rho, theta, theta): -1.0*rho, (theta, rho, theta): 1.0/rho, (theta, theta, rho): 1.0 / rho}
        print(true_values)
        assert cs.values == true_values


if __name__ == '__main__':
    unittest.main()



import numpy as np
import sympy as sp
from sympy.matrices import Matrix

class NotContainedSymbolException(Exception):
    ...


class IndexObject:

    def __init__(self, head, coords, values, index_pos=None):
        self.head = head
        self.coords = coords
        self.values = values
        if index_pos is None:
            index_pos = len(list(values.keys())[0]) * [0]
        self.index_pos = index_pos

        for key in values:
            self.check_valid_index(key)

    def display(self):
        display_parts = []
        for k, v in self.values.items():
            index_strings = []
            for index_pos, k_i in zip(self.index_pos, k):
                index_location = "_" if index_pos == 0 else "^"
                index_strings.append(index_location + sp.latex(k_i))
            display_parts.append(self.head + "{}".join(index_strings) + " = " + str(v))
        dollar = "$$"
        return dollar + r" \\ ".join(display_parts) + dollar

    def __getitem__(self, item):
        self.check_valid_index(item)
        return self.values.get(item, 0)

    def __setitem__(self, item, value):
        self.check_valid_index(item)
        if value != 0:
            self.values[item] = value
        else:
            if item in self.values:
                del self.values[item]

    def check_valid_index(self, index_list):
        assert len(index_list) == len(self.index_pos), "Values for all indices should be specified."
        for symbol in index_list:
            if symbol not in self.coords:
                raise NotContainedSymbolException


class Manifold:

    def __init__(self, coords, metric_values):
        self.coords = coords
        if type(metric_values) == dict:
            values = metric_values

        elif type(metric_values) == Matrix:
            values = self.matrix_to_dict(metric_values)
        else:
            raise TypeError()

        self.metric = Tensor(
            head="g",
            manifold=self,
            index_pos=(0, 0),
            values=values)

        self.metric_inv = Tensor(self.metric.head,
                                 self, (1, 1),
                                 self.matrix_to_dict(self.metric_to_matrix().inv()))

    def matrix_to_dict(self, metric_values):
        values = {}
        size = sp.shape(metric_values)[0]
        for i in range(size):
            for j in range(size):
                if metric_values[i, j] != 0:
                    values[self.coords[i], self.coords[j]] = metric_values[i, j]
        return values

    def metric_to_matrix(self):
        size = len(self.coords)
        mat = Matrix(np.zeros([size, size]))

        for i in range(size):
            for j in range(size):
                mat[i, j] = self.metric[self.coords[i], self.coords[j]]

        return mat


class Tensor(IndexObject):

    def __init__(self, head, manifold, index_pos, values):
        self.manifold = manifold
        super(Tensor, self).__init__(
            head,
            manifold.coords,
            values,
            index_pos=index_pos)


class ChristoffelSymbol(IndexObject):
    def __init__(self, manifold):
        self.manifold = manifold
        g = self.manifold.metric
        coords = self.manifold.coords
        g_inv = self.manifold.metric_inv

        super(ChristoffelSymbol, self).__init__(r"\Gamma", manifold.coords, {}, index_pos=(1, 0, 0))
        for i_mu in coords:
            for i_alpha in coords:
                for i_beta in coords:
                    for i_sigma in coords:
                        self[(i_mu, i_alpha, i_beta)] = self[(i_mu, i_alpha, i_beta)] + 0.5 * g_inv[
                            i_mu, i_sigma] * (sp.diff(g[i_alpha, i_sigma], i_beta)
                                              + sp.diff(g[i_sigma, i_beta], i_alpha)
                                              - sp.diff(g[i_alpha, i_beta], i_sigma))



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from setuptools import setup, find_packages

name = 'tensor_algebra_dont_use'

setup(
    name=name,
    author='Dojo_team_1',
    version='0.0.1',
    packages=find_packages(),
    package_dir={name: name},
    include_package_data=True,
    license='MIT',
    description='bla',
    long_description='bla bla',
    install_requires=['sympy>=1.8',
                      'numpy>=1.17.0',
                      ],
    python_requires=">=3.6"
)